//***************************************************************
// Filename: 	CLBRead.c
//
// Brief:	This program reads a CLB structure 
//		from file 
//
// Author:	Sverre Hamre, 2009, NTNU
//
//***************************************************************
#include <stdio.h>
#include <stdlib.h>
#include <stdint.h>
#include <string.h>
#include "CLBRead.h"

//for printing data to screen at x86 arcitectures
uint32_t endian(uint32_t input){
	return((input&0xff)<<24)+((input&0xff00)<<8)+((input&0xff0000)>>8)+((input>>24)&0xff);
}

uint32_t locateFrameStart(FILE* infile)
{
  uint32_t word=0;
  uint32_t configurationWords;
  uint32_t i=0, j=0;
  uint8_t  byte=0;
  uint32_t work=1;


  rewind(infile);
  while(work){
    fread(&byte, sizeof(uint8_t), 1, infile);
    if(byte == 0xFF){
      j++;
      if(j == 4)                    // found dummy word start reading 32 bit from here on.
	work = 0;
    }else{
      j=0;
    }
    i++;
  }

  // find synchronization word, should be just after the dummy word.
  while(word != SYNCHRONIZATION_WORD){
    fread(&word, sizeof(uint32_t), 1, infile);
    i += 4;
  }
  DEBUG(printf("locateFrameStart: Synch word = %x \n", endian(word)));
  
  // find the Frame Data Input Register access.
  while(word != FDRI_ADR){
    fread(&word, sizeof(uint32_t), 1, infile);
    i += 4;
  }
  DEBUG(printf("locateFrameStart: FDRI word = %x \n", endian(word)));
  
  // read in the data going to the FDRI.
  fread(&word, sizeof(uint32_t), 1, infile);
  DEBUG(printf("locateFrameStart: word = %x \n", endian(word)));
  i += 4;

  // Check if type 1 or type 2 packet header
  if((word & 0xe0) == 0b01000000){
    DEBUG(printf("locateFrameStart: Type2 header detected\n"));
    configurationWords = endian(word) & 0x07FFFFFF;
    DEBUG(printf("locateFrameStart: Word count: %d \n", configurationWords));
    DEBUG(printf("locateFrameStart: Number of configuration bits: %d \n", configurationWords*32));
  }
  else if((word & 0xe0) == 0b00100000){
    DEBUG(printf("locateFrameStart: Type1 header detected\n"));
    configurationWords = endian(word) & 0x000003FF;
    DEBUG(printf("locateFrameStart: Word count: %d \n", configurationWords));
    DEBUG(printf("locateFrameStart: Number of configuration bits: %d \n", configurationWords*32));
  }

  DEBUG(printf("locateFrameStart: Configuration bytes = %d \n", i));

  rewind(infile);
  return i;
}

  /* First deside from CLBnr which frame to start from configuration frames start at center
  *  of fpga and go to right then goes a row up, when finished with top half of fpga next 
  *  frame is center under left bottom, frames count to left then goes row down.
  *
  *  11111111111     
  *  00000000000     start row
  *  22222222222
  *  33333333333     end row for virtex-4 fx12.
  *
  *  Columns within a row:
  *  
  *  IOB | CLB | IOB | CLK | CLB | DSP | CLB | IOB | overh | BRAMI | overh | BRAM | overh |
  *   1  | 12  |  1  |  1  |  4  |  1  |  8  |  1  |   1   |   3   |   1   |   3  |   1   |
  *
  *  Bitfile is aranged so IOB, CLB, CLK and DSP is first 3248 frames for fx12 and then 
  *  comes BRAM data 352 frames.
  *
  *  Frames pr column by type for Virtex-4FX12:
  *  IOB       30
  *  CLB       22
  *  CLK        2
  *  DSP       21
  *  BRAMI     20
  *  BRAM      64
  *  overhead   2
  */
uint32_t locateFrameFromCLBnr(uint16_t CLBnr){

  uint32_t frame;
  uint32_t CLBcol;

  // First decide row.
  if(CLBnr < 384){
    frame = 644;
    DEBUG(printf("locateFrameFromCLBnr: frame: %d \n", frame));
  } else if((384 <= CLBnr) && (CLBnr < 768)){
    frame = 0;
    DEBUG(printf("locateFrameFromCLBnr: frame: %d \n", frame));
  } else if((768 <= CLBnr) && (CLBnr < 1152)){
    frame = 1288;
    DEBUG(printf("locateFrameFromCLBnr: frame: %d \n", frame));
  } else if((1152 <= CLBnr) && (CLBnr < 1535)){
    frame = 1932;
    DEBUG(printf("locateFrameFromCLBnr: frame: %d \n", frame));
  } else {
    // Out of range
    printf("\nCLBnr %d, hex %X\n", CLBnr, CLBnr);
    printf("ERROR: locate Frame From CLB nr: CLBnr is out of range\n\n");
    return 0;
  }

  // Decide start frame within row.
  CLBcol = CLBnr % CLB_PR_ROW;
  DEBUG(printf("locateFrameFromCLBnr: CLBcol: %d \n", CLBcol));
  if(CLBcol < 12){
    frame += FRAMES_PR_IOB + (CLBcol*FRAMES_PR_CLB);
    DEBUG(printf("locateFrameFromCLBnr: frame: %d \n", frame));
  } else if((12 <= CLBcol) && (CLBcol < 16)){
    frame += (FRAMES_PR_IOB*2) + FRAMES_PR_CLK + (CLBcol*FRAMES_PR_CLB);
    DEBUG(printf("locateFrameFromCLBnr: frame: %d \n", frame));
  } else if((16 <= CLBcol) && (CLBcol < 24)){
    frame += (FRAMES_PR_IOB*2) + FRAMES_PR_CLK + FRAMES_PR_DSP + (CLBcol*FRAMES_PR_CLB);
    DEBUG(printf("locateFrameFromCLBnr: frame: %d \n", frame));
  } else {
    // Out of range
    printf("\nCLBnr %d, hex %X\n", CLBcol, CLBcol);
    printf("ERROR: locate Frame From CLB nr: CLBcol is out of range\n\n");
    return 0;
  }

  return (frame * FRAME_LENGTH_BYTES);
}

/*
 *  Reads frames from input file and writes to a outputfile.
 *
 *  One column of CLBs or more gets read out, one column is 16 CLB high.
 *  One column of CLBs is 22 frames. 
 *
 *  Output file will just have the frames read out.
 */
int read_frames(CLBinputs_t *inputs){

  int ret = 0;
  uint32_t frames;
  uint32_t frameStart;
  uint32_t *frameBuffer;
  FILE *infile, *outfile;
  uint16_t startCLB, endCLB;

  DEBUG(printf("read_frame: acessing\n"));
  // open a buffer for 1 frame.
  frameBuffer = (uint32_t*)malloc(sizeof(uint32_t) * FRAME_LENGTH_32BIT_WORDS 
				  * ((((inputs->endCLB % 24)+1) - (inputs->startCLB % 24)) * FRAMES_PR_CLB));
  if(frameBuffer == NULL){
      if(inputs->verbose) printf("ERROR: read_frame: Unable to alocate frameBuffer memory\n");
  }else{
    if(inputs->verbose) printf("read_frame: Memory alocated\n");
    
    // Open input/output files.
    infile = fopen(inputs->infilename, "r");
    outfile = fopen(inputs->outfilename, "w");

    if((infile == NULL) || (outfile == NULL)){
      if(infile == NULL){
	if(inputs->verbose) printf("read_frame: Unable to open ");   
	if(inputs->verbose) printf(inputs->infilename);
	if(inputs->verbose) printf("\n");      
      }
      if(outfile == NULL){
	if(inputs->verbose) printf("read_frame: Unable to open ");   
	if(inputs->verbose) printf(inputs->outfilename);
	if(inputs->verbose) printf("\n");      
      }
    }else{

      // Find start of first frame. Jump to CLB frames.
      frameStart = locateFrameStart(infile);
      frameStart += locateFrameFromCLBnr(inputs->startCLB);    // Jump to desired frame... calculated from CLB nr.

      // Number of frames to be read out.
      if( (inputs->startCLB % 24) >  (inputs->endCLB % 24) ){
	printf("read_frame: error start CLB higher than end CLB\n");
      }else{
	frames = (((inputs->endCLB % 24)+1) - (inputs->startCLB % 24)) * FRAMES_PR_CLB;
	DEBUG(printf("read_frame: frames to read out: %d\n", frames));


	DEBUG(printf("read_frame: frameStart: %d\n", frameStart));
	fseek(infile, frameStart, SEEK_SET);

	
	startCLB = inputs->startCLB % CLB_PR_ROW;
	endCLB   = inputs->endCLB % CLB_PR_ROW;
	DEBUG(printf("read_frame: Read out CLBs startCLB: %d endCLB: %d \n", startCLB, endCLB));
	// Read out CLBs
	/* Not pretty code, this should be cleaned up for easier readability*/
	if(startCLB < 12){
	  DEBUG(printf("read_frame: startCLB < 12 \n"));
	  if(endCLB < 12){
	    DEBUG(printf("read_frame: endCLB < 12\n"));
	    fread(frameBuffer, sizeof(uint32_t), (frames * FRAME_LENGTH_32BIT_WORDS), infile);
	    fwrite(frameBuffer, sizeof(uint32_t), (frames * FRAME_LENGTH_32BIT_WORDS), outfile);
	  } else {
	    DEBUG(printf("read_frame: endCLB > 12 \n"));
	    printf("BUGYBUGYBUGBUGBUG BUGBUG FIX\n");
	    // read first frames to second IOB
	    fread(frameBuffer, sizeof(uint32_t), ((12 - inputs->startCLB) * FRAME_LENGTH_32BIT_WORDS), infile);
	    fwrite(frameBuffer, sizeof(uint32_t), ((12 - inputs->startCLB) * FRAME_LENGTH_32BIT_WORDS), outfile);
	    // Jump over IOB and CLK
	    fseek(infile, ((FRAMES_PR_IOB + FRAMES_PR_CLK) * FRAME_LENGTH_BYTES), SEEK_CUR);
	  }

	  if((12 <= endCLB) && (endCLB < 16)){
	    // read out rest of frames.
	    DEBUG(printf("read_frame: startCLB >= 12 endCLB < 16\n"));
	    printf("BUGYBUGYBUGBUGBUG BUGBUG FIX\n");
	    fread(frameBuffer, sizeof(uint32_t), 
		  ((frames - (12 - startCLB)) * FRAME_LENGTH_32BIT_WORDS), infile);
	    fwrite(frameBuffer, sizeof(uint32_t), 
		   ((frames - (12 - startCLB)) * FRAME_LENGTH_32BIT_WORDS), outfile);
	  } else if (16 <= endCLB) {
	    DEBUG(printf("read_frame: endCLB > 16\n"));
	    printf("BUGYBUGYBUGBUGBUG BUGBUG FIX\n");
	    // read frames after IOB and CLK to DSP, this is 4 CLBs.
	    fread(frameBuffer, sizeof(uint32_t), 
		  (4 * FRAMES_PR_CLB * FRAME_LENGTH_32BIT_WORDS), infile);
	    fwrite(frameBuffer, sizeof(uint32_t), 
		   (4 * FRAMES_PR_CLB * FRAME_LENGTH_32BIT_WORDS), outfile);
	    // Jump over DSP frames.
	    fseek(infile, (FRAMES_PR_DSP * FRAME_LENGTH_BYTES), SEEK_CUR);
	    // read out frames after DSP frames.
	    fread(frameBuffer, sizeof(uint32_t), 
		  ((endCLB - 15) * FRAME_LENGTH_32BIT_WORDS), infile);
	    fwrite(frameBuffer, sizeof(uint32_t), 
		   ((endCLB - 15) * FRAME_LENGTH_32BIT_WORDS), outfile);
	  }

	} else if((12 <= startCLB) && (startCLB < 16)){
	  DEBUG(printf("read_frame: startCLB >= 12 startCLB < 16\n"));
	  printf("BUGYBUGYBUGBUGBUG BUGBUG FIX\n");
	  if(endCLB < 16){
	    DEBUG(printf("read_frame: endCLB < 16\n"));
	    // read out rest of frames.
	    fread(frameBuffer, sizeof(uint32_t), 
		  ((endCLB - 11) * FRAME_LENGTH_32BIT_WORDS), infile);
	    fwrite(frameBuffer, sizeof(uint32_t), 
		   ((endCLB -11) * FRAME_LENGTH_32BIT_WORDS), outfile);
	  } else {
	    DEBUG(printf("read_frame: endCLB > 16\n"));
	    // read frames after IOB and CLK to DSP, this is 4 CLBs.
	    fread(frameBuffer, sizeof(uint32_t), 
		  (4 * FRAMES_PR_CLB * FRAME_LENGTH_32BIT_WORDS), infile);
	    fwrite(frameBuffer, sizeof(uint32_t), 
		   (4 * FRAMES_PR_CLB * FRAME_LENGTH_32BIT_WORDS), outfile);
	    // Jump over DSP frames.
	    fseek(infile, (FRAMES_PR_DSP * FRAME_LENGTH_BYTES), SEEK_CUR);
	    // read out frames after DSP frames.
	    fread(frameBuffer, sizeof(uint32_t), 
		  ((endCLB - 15) * FRAME_LENGTH_32BIT_WORDS), infile);
	    fwrite(frameBuffer, sizeof(uint32_t), 
		   ((endCLB - 15) * FRAME_LENGTH_32BIT_WORDS), outfile);
	  }
	  
	} else if((16 <= startCLB) && (startCLB < 24)){
	  // read out frames after DSP frames.
	  DEBUG(printf("read_frame: startCLB >= 16 startCLB < 24\n"));
	  fread(frameBuffer, sizeof(uint32_t), (frames * FRAME_LENGTH_32BIT_WORDS), infile);
	  fwrite(frameBuffer, sizeof(uint32_t), (frames * FRAME_LENGTH_32BIT_WORDS), outfile);
	}

	// Read out connections in BRAM DSP if necessary.
	printf("warning BRAM/DSP connections read out not implemented\n");


      }
      // close files.
      DEBUG(printf("read_frame: Closing files\n"));
      fclose(infile);
      fclose(outfile);
    }
    // close buffer.
    free(frameBuffer);
  }

  return ret;

}

/*
 *  Writes frames from input file and writes to a outputfile.
 *
 *  One column of CLBs or more gets read out, one column is 16 CLB high.
 *  One column of CLBs is 22 frames. 
 *
 *  Inputfile will just have the frames to write, outfile will be a full
 *  Configuration bitstream.
 */
int write_frames(CLBinputs_t *inputs){
  int ret = 0;
  uint32_t frames;
  uint32_t frameStart;
  uint32_t *frameBuffer;
  FILE *infile, *outfile;
  uint16_t startCLB, endCLB;


  DEBUG(printf("write_frame: acessing\n"));
  // open a buffer for frames.
  frameBuffer = (uint32_t*)malloc(sizeof(uint32_t) * FRAME_LENGTH_32BIT_WORDS 
				  * ((((inputs->endCLB % 24)+1) - (inputs->startCLB % 24)) * FRAMES_PR_CLB));
  if(frameBuffer == NULL){
      if(inputs->verbose) printf("ERROR: read_frame: Unable to alocate frameBuffer memory\n");
  }else{
    if(inputs->verbose) printf("write_frame: Memory alocated\n");

    // Open input/output files.
    infile = fopen(inputs->infilename, "r");
    outfile = fopen(inputs->outfilename, "r+");
    

    if((infile == NULL) || (outfile == NULL)){
      if(infile == NULL){
	if(inputs->verbose) printf("write_frame: Unable to open ");   
	if(inputs->verbose) printf(inputs->infilename);
	if(inputs->verbose) printf("\n");      
      }
      if(outfile == NULL){
	if(inputs->verbose) printf("write_frame: Unable to open ");   
	if(inputs->verbose) printf(inputs->outfilename);
	if(inputs->verbose) printf("\n");      
      }
    }else{
      // Find start of first frame. Jump to CLB frames.
      frameStart = locateFrameStart(outfile);
      frameStart += locateFrameFromCLBnr(inputs->startCLB);    // Jump to desired frame... calculated from CLB nr.

      // Number of frames to be read out.
      if( (inputs->startCLB % CLB_PR_ROW) >  (inputs->endCLB % CLB_PR_ROW) ){
	printf("write_frame: error start CLB higher than end CLB\n");
      }else{
	frames = (((inputs->endCLB % CLB_PR_ROW)+1) - (inputs->startCLB % CLB_PR_ROW)) * FRAMES_PR_CLB;
	DEBUG(printf("write_frame: frames to write out: %d\n", frames));

	
	startCLB = inputs->startCLB % CLB_PR_ROW;
	endCLB = inputs->endCLB % CLB_PR_ROW;
	DEBUG(printf("write_frame: startCLB: %d endCLB: %d\n", startCLB, endCLB));

	// simple implementation will error when CLBs cross logic elements.
	if( ((startCLB < 12) && (endCLB < 12))
	    || ((startCLB >= 12) && (startCLB < 16) && (endCLB >= 12) && (endCLB < 16))
	    || ((startCLB >= 16) && (startCLB < 24) && (endCLB >= 16) && (endCLB < 24))){
	  
	  // read out frames.
	  fread(frameBuffer, sizeof(uint32_t), (frames * FRAME_LENGTH_32BIT_WORDS), infile);

	  // write frames. that are from infile
	  DEBUG(printf("write_frame: Writing frames, frameStart: %d\n", frameStart));
	  fseek(outfile, frameStart, SEEK_SET);
	  fwrite(frameBuffer, sizeof(uint32_t), (frames * FRAME_LENGTH_32BIT_WORDS), outfile);

	
	}else{
	  printf("write_clb: ERROR cross logick element frame write yet not implemented\n");
	}
      }
      // close files.
      DEBUG(printf("write_frame: Closing files\n"));
      fclose(infile);
      fclose(outfile);
    }
    // close buffer.
    free(frameBuffer);
  }

  return ret;
}


/*
 * prints out one CLB from start CLB at top left to end CLB
 * bottom right.
 *
 * used for debuging.
 */
void print_CLB(CLBinputs_t *inputs){
  uint32_t frameStart;
  FILE *infile;
  uint16_t CLBoffset;
  uint16_t i, j;
  uint8_t CLBdata[BYTES_PR_CLB];

  // Open input/output files.
  infile = fopen(inputs->infilename, "r");

  if(infile == NULL){
    printf("print_CLB: error opening file\n");
    return;
  }

  // Find start of first frame. Jump to CLB frames.
  frameStart = locateFrameStart(infile);
  frameStart += locateFrameFromCLBnr(inputs->startCLB);    // Jump to desired frame... calculated from CLB nr.

  // Jump to frame
  DEBUG(printf("print_CLB: frameStart: %d\n", frameStart));
  fseek(infile, frameStart, SEEK_SET);

  // offset is different from over center start of frame is always closest clb to center of FPGA
  if(inputs->startCLB < 768){
    CLBoffset = (inputs->startCLB  / CLB_PR_ROW) % CLB_PR_FRAME;
    CLBoffset = CLB_PR_FRAME - (CLBoffset + 1);
    DEBUG(printf("print_CLB: CLBoffset: %d\n", CLBoffset));
  } else {
    CLBoffset = (inputs->startCLB / CLB_PR_ROW) % CLB_PR_FRAME;
    DEBUG(printf("print_CLB: CLBoffset: %d\n", CLBoffset));
  }

  if(CLBoffset >= 8){
    DEBUG(printf("print_CLB: CLBoffset: %d\n", (BYTES_CENTER_WORD+(CLBoffset*BYTES_PR_CLB))));
    fseek(infile, BYTES_CENTER_WORD+(CLBoffset*BYTES_PR_CLB), SEEK_CUR);       // Jump to corrcet CLB.	
  }else{
    DEBUG(printf("print_CLB: CLBoffset: %d\n", (CLBoffset*BYTES_PR_CLB)));
    fseek(infile, CLBoffset*BYTES_PR_CLB, SEEK_CUR);       // Jump to corrcet CLB.	
  }

  
  // Print out the CLB data.
  printf("CLB %d data:\n", inputs->startCLB);
  for(i=0;i<FRAMES_PR_CLB;i++){
    fread(CLBdata, BYTES_PR_CLB, 1, infile);
    for(j=0;j<BYTES_PR_CLB;j++){
      printf("%X ", CLBdata[j]);
    }
    fseek(infile, FRAME_LENGTH_BYTES - BYTES_PR_CLB, SEEK_CUR);       // Jump to next frame.
    printf("\n");
  }

  // Close file.
  fclose(infile);
  return;    
}
