#include <stdio.h>
#include <stdlib.h>
#include <stdint.h>
#include <fcntl.h>
#include <unistd.h>
#include <string.h>

#include "icap_write.h"

#define ICAP "/var/icap" // Memory mapped FPGA

/*
 *  Setup hwicap to write out frames.
 */
int write_header(int device, uint32_t row, uint32_t column, uint32_t frame, int frames){
  uint32_t data[19];
  int ret=0;
  uint32_t address;
  uint32_t packet;
  uint32_t headerWords;
  uint32_t totalWords;

  // Dummy, synch  
  data[0] = DUMMY_PACKET;
  data[1] = SYNC_PACKET;
  data[2] = NOP_PACKET;    //Flush
  data[3] = NOP_PACKET;    //Flush

  // Reset CRC
  data[4] = CMD_PACKET;
  data[5] = RCRC_CMD;
  data[6] = NOP_PACKET;    //Flush
  data[7] = NOP_PACKET;    //Flush

  // Id register
  data[8] = IDCODE_PACKET;
  data[9] = V4FX12ID;      //Device ID code;

  // Top/bottom bit, block type, row address, column, frame
  address = (0<<22) | (0<<19) | (row<<14) | (column<<6) | frame;

  // Setup FAR
  data[10] = FAR_PACKET;
  data[11] = address;       // address (top, CLB, row, column, minor)

  // Setup CMD write configuration.
  data[12] = CMD_PACKET;
  data[13] = WCFG_CMD;      // Type1 write  words to FDRI
  data[14] = NOP_PACKET;    //Flush

  totalWords = ((frames+1)*WORDS_PR_FRAME);     //Frames + dumy frame times
						//words pr frame.
  if(totalWords < XHI_TYPE_1_PACKET_MAX_WORDS){
    // Create Type 1 packet
    packet = (1<<TYPE1) | (1<<OP_WRITE) | (FDRI<<REG_ADDR) | totalWords;
    data[15] = packet;
    
    headerWords = 16;
  }else{
    // Create Type 2 packet
    packet = (1<<TYPE1) | (1<<OP_WRITE) | (FDRI<<REG_ADDR);
    data[15] = packet;
    
    packet = (1<<TYPE2) | (1<<OP_WRITE) | totalWords;
    data[16] = packet;
  
    headerWords = 17;
  }

  // Write out header.
  ret = write(device, data, sizeof(uint32_t)*headerWords);   // write out data
							     // to device.
  return ret;
}


int write_frames(int device, uint32_t frames, char* filename,uint32_t frameOffset){
  int ret=0;
  void *frameBuffer;
  FILE *inFile;

  // allocate memmory for frames.
  frameBuffer = malloc(sizeof(uint32_t)*WORDS_PR_FRAME*(frames+1));
  if(frameBuffer == NULL){
    printf("Error alocating frame buffer\n");
    return -1;   // ERROR
  }
  
  // open a file to read frames from. 
  inFile = fopen(filename, "r");    // Read frames from this file.
  if(inFile == NULL){
    printf("error opening file\n");
    free(frameBuffer);
    return -1;    // ERROR
  }

  // Go to correct place in file.
  if(fseek(inFile, frameOffset*WORDS_PR_FRAME*sizeof(uint32_t), SEEK_SET) < 0){
    printf("error seting filepointer offset\n");
    free(frameBuffer);
    fclose(inFile);
    return -1; // ERROR;
  }

  // Read frames from file and insert in buffer.
  if(fread(frameBuffer, WORDS_PR_FRAME*sizeof(uint32_t), (frames) , inFile) < 0){
    printf("error reading from file\n");
    ret = -1;
  }else{
    // Write frames + PAD frame.
    // As the memory is not initialized the PAD frame here can be anything.
    // Is this a BUG? should not mather as it is just for flushing the framebuffer.
    if(write(device, frameBuffer, sizeof(uint32_t)*WORDS_PR_FRAME*(frames+1)) < 0){
      printf("Error writing frames to device\n");
      ret = -1;
    }
  }

  // Clean up.
  free(frameBuffer);
  fclose(inFile);

  return ret;  
}


int write_postconfig(int device){
  uint32_t data[5];
  int ret=0;

  // CRC
  data[0] = CMD_PACKET;
  data[1] = XHI_DISABLED_AUTO_CRC;

  // Desync
  data[2] = DESYNC_CMD;
  data[3] = NOP_PACKET;
  data[4] = NOP_PACKET;

  ret = write(device, data, 4*5);   // write;

  return ret;
}

uint32_t get_status(int device){
  uint32_t data[7];
  uint32_t status;

  // Dummy, synch  
  data[0] = DUMMY_PACKET;
  data[1] = SYNC_PACKET;
  data[2] = NOP_PACKET;    //Flush
  data[3] = NOP_PACKET;    //Flush

  // Read out status, one word.
  data[4] = (1<<TYPE1) | (1<<OP_READ) | (STAT<<REG_ADDR) | 1; 
  data[5] = NOP_PACKET;
  data[6] = NOP_PACKET;

  printf("Stat acces word %X\n",data[4]);


  if(write(device, data, sizeof(uint32_t)*7) == 0){
    printf("error writing get status\n");
    return 0; //error
  }

  // Get status.
  if(read(device, &status, sizeof(uint32_t)) == 0){
    printf("error reading status\n");
    return 0; //error
  }

  // Desync
  data[0] = DESYNC_CMD;
  data[1] = NOP_PACKET;
  data[2] = NOP_PACKET;

  if(write(device, data, sizeof(uint32_t)*3) == 0){   // write;
    printf("error desyncing get status\n");
    return 0; //error
  }

  return status;
}


int main(int argc, char * argv[])
{
  int handlemem;
  int i;
  int numArgs=0;
  uint32_t frames;
  uint32_t status;
  char *infilename;
  


  if(argc != 1){
    for(i=1;i<argc;i++){
      if(*argv[i] == '-'){
	if(*(argv[i]+1) == 'h'){
	  printf("icap_write -i [filename] -f [frames]\n");
	  printf("This program has the address hardcoded inn.\n");
	  printf("Frames will be written to CLB 21 and on, on top row.\n"); 
	  break;          //quit after printing help. 
	}
	else if(strcmp((argv[i]+1), "i") == 0){
	  infilename = argv[i]+3;     // store pointer to input filename
	  numArgs++;
	}
	else if(strcmp((argv[i]+1), "f") == 0){
	  frames = atoi(argv[i]+3);         // Frames to be writen in.
	  printf("Frames: %d\n",frames);
	  numArgs++;
	}
      }
    }
    
    if(numArgs == 2){
      printf("Access Hardware \n");
      handlemem = open(ICAP,O_RDWR);
      if(handlemem == -1){
	printf("FPGA Memory could not be accessed\n");
	return 0;
      }else{
	printf("FPGA Memory accessing\n");
      }
      
      // Address hardcoded in, row 2(=1) from center, CLB 21(=25), frame 0.
      if(write_header(handlemem, 1, 25, 0, frames) < 0){
	printf("Error writing header\n");
      }else{
	printf("Writing frames\n");
	if(write_frames(handlemem, frames, infilename, 0) < 0){
	  printf("Error Writing frames\n");
	  
	}else{
	  printf("Finished writing frames\n");
	  printf("Postconfig\n");
	  
	  if(write_postconfig(handlemem) < 0){
	    printf("error writing postconfig\n");
	  }
	}
      }
    
      printf("Getting Status of FPGA\n");
      status = get_status(handlemem);
      if(status == 0){
	printf("error getting status\n");
      }else{
	printf("Status: %X\n", status);
      }
    
    
      // Clean up.
      printf("closing device\n");
      if(close(handlemem) != 0){
	perror("closing driver");
      }
    }
  } else {
    printf("Execute with -h for help\n");
  }
  
  return 0;
}
