--------------------------------------------------------------
---------------RECONFIGURABLE NFSM CLONE-------------------
--Top module for the reconfigurable nfsm clone
--State machine can contain a variable number of states, default 4, simulating an NFSM
--State machine contains a configuration memory module that can be filled with a transition table
--The transition table is used by the NFSM module that contains the states, as flip-flops, and combinatorial that checks the transitions

--author: Tormod Heimark
--date:   02/03/2015
--------------------------------------------------------------

library IEEE;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;

library work;
use work.check_config_package.all; 

entity nfsm_clone is 
generic (
			INPUT_SIZE : natural :=  1;
			NR_STATES : natural  :=  4;
			MEM_SIZE : natural   :=  5;
			NUM_REG : natural    := 25
			);
port (	clk     : in std_logic;
		   dataIn  : in std_logic_vector(15 downto 0);
		   dataOut : out std_logic_vector(7 downto 0)
	);
end nfsm_clone;

architecture arch of nfsm_clone is

------------------INITIATE MODULES-----------------	
	component nfsm
		 generic (
		  INPUT_SIZE : natural :=  1;
        NUM_REG    : natural := 25;
		  MEM_SIZE   : natural :=  5;
		  NR_STATES  : natural :=  4
		);
		port (
			input           : in  std_logic_vector(INPUT_SIZE-1 downto 0); 
			clk             : in  std_logic;
			reset           : in  std_logic;
			hold			    : in  std_logic;
			paralell_in     : in  std_logic_vector(NUM_REG*MEM_SIZE-1 downto 0);
			force_states_en : in std_logic;
			force_states    : in std_logic_vector(NR_STATES-1 downto 0);
			
			delete_state : out std_logic;
			copy_state   : out std_logic;
			accept       : out std_logic
			
		);
	end component;
	
	component configuration_memory
		generic(
			MEM_SIZE : natural  := 5;
			NUM_REG  : natural  := 20
	   );
	port (
		clk	     : in std_logic;
		shift_i    : in std_logic_vector (MEM_SIZE-1 downto 0);
		write_en   : in std_logic;
		
		paralell_o : out std_logic_vector(MEM_SIZE*NUM_REG-1 downto 0) 
	);
	end component;
	------------------------------------------------------------------------------	
	
	----------------SIGNALS TO CONFIGURATION MEMORY-------------------------------
	signal shift_in_to_conf            : std_logic_vector (MEM_SIZE-1 downto 0);
	signal paralell_out_to_comb        : std_logic_vector(MEM_SIZE*NUM_REG-1 downto 0); 
	signal enable_writing_to_conf      : std_logic;
	------------------------------------------------------------------------------
	
	----------------SIGNALS CONNECTED TO NFSM-----------------------------------------------
	signal reset_ff, next_reset_ff : std_logic;
	signal hold_ff,next_hold_ff    : std_logic;
	signal input_to_comb           : std_logic_vector(INPUT_SIZE-1 downto 0);
	signal force_states_comb       : std_logic_vector(NR_STATES-1 downto 0);
	signal accept_from_comb        : std_logic;
	signal copy_state_from_comb    : std_logic;
	signal delete_signal_comb      : std_logic;
	signal force_en                : std_logic;
	-----------------------------------------------------------------------------
	 
	----------------SIGNALS FOR STATE MACHINE------------------------------------
	type operational_state is (IDLE,CONFIGURE,OPERATE); 
	signal current_state, next_state             : operational_state;
	signal configure_count, next_configure_count : integer range 0 to NUM_REG := 0;
	-----------------------------------------------------------------------------
	
	-----------------------------------------------------------------------------
	--STATES:
		--IDLE      : NFSM deactivated, no vectors on conf_mem, waiting for enable on dataIn(6)
		--CONFIGURE : NFSM deactivated, shifting in vectors to conf_mem, duration set by counter configure count
		--OPERATE   : NFSM activated is evaluating input and current state with vectors in conf_mem
		--DELETED
	-----------------------------------------------------------------------------
	
	-------------------------PINS------------------------------------------------
	-- dataIn(7-0)    = input from form backend, shifted in
	-- dataIn(8)      = configure enable, 1 active pulse signals configure vektors on next cycle
	-- dataIn(9)      = force states enable signal
	-- dataIN(10)     = valid input
	-- dataIn(11)     = reset, active high
	
	-- dataOut(0)     = accept output from NFSM
	-- dataOut(1)     = copy the state, 
	-- dataOut(2)     = deleted state
	-----------------------------------------------------------------------------
begin
	
	
	
	NFSM0 : nfsm 
	generic map(
		INPUT_SIZE => INPUT_SIZE,
		NUM_REG    => NUM_REG,
		MEM_SIZE   => MEM_SIZE,
		NR_STATES  => NR_STATES
	)
	port map ( 
		input           => input_to_comb,
		paralell_in     => paralell_out_to_comb,
		force_states_en => force_en,
		force_states    => force_states_comb,
		accept          => accept_from_comb,
		copy_state      => copy_state_from_comb,
		delete_state    => delete_signal_comb,
		reset           => reset_ff,
		clk             => clk,
		hold            => hold_ff
		);
												 
	CONF_MEM : configuration_memory 	
	generic map ( 
		MEM_SIZE => MEM_SIZE,
		NUM_REG => NUM_REG
	)
	port map ( clk => clk,
		shift_i => shift_in_to_conf,
		write_en => enable_writing_to_conf,
		paralell_o => paralell_out_to_comb
	);
																
	-------SOME CONNECTIONS FOR IN AND OUT---------
	--input_to_comb <= dataIn(INPUT_SIZE-1 downto 0);
	
	force_en          <= dataIn(9);
	force_states_comb <= dataIn(NR_STATES-1 downto 0);
	
	dataOut(5 downto 3)   <= (others=>'0');
	dataOut(6)            <= dataIn(5) or dataIn(4) or dataIn(6) or dataIn(7);
	dataOut(7)            <= dataIn(12) or dataIn(13) or dataIn(14) or dataIn(15);
	dataOut(1)            <= copy_state_from_comb;
	dataOut(0)            <= accept_from_comb;
	dataOut(2)            <= delete_signal_comb;
	shift_in_to_conf      <= dataIn(MEM_SIZE-1 downto 0);

	-----------------------------------------------


	process(clk,dataIn)
	begin
		if rising_edge(clk) then 
			if ( dataIn(11) = '1') then -- Reset the state machine if reset is high
				current_state <= IDLE;
				configure_count <= 0;
				hold_ff <= '0';
				reset_ff <= '1';
				input_to_comb <= (others=>'0');
			else
				current_state <= next_state;
				configure_count <= next_configure_count;
				hold_ff <= next_hold_ff;
				reset_ff <= next_reset_ff;
				if ( current_state = OPERATE) then
					input_to_comb <= dataIn(INPUT_SIZE-1 downto 0);
				else
					input_to_comb <= (others=>'0');
				end if;

			end if;
		end if;
	end process;		
	
	clone0 : process ( dataIn, 
						current_state, 
						configure_count
						)
		begin
		
			
			----------------STATE MACHINE OPERATION--------
			case current_state is
				when IDLE =>                  
					if ( dataIn(8) = '1') then --configure enable from backend, go to CONFIGURE
						next_state <= CONFIGURE;
					else
						next_state <= IDLE;
					end if;
					
					next_configure_count <= 0;
					next_reset_ff <= '1';
					enable_writing_to_conf <= '0';
					next_hold_ff <= '0';
				when CONFIGURE =>
					if (configure_count < NUM_REG) then
						enable_writing_to_conf <= '1';
						next_configure_count <= configure_count+1;
						next_state <= CONFIGURE;
					else
						next_state <= OPERATE;
						enable_writing_to_conf <= '0';
						next_configure_count <= 0;
					end if;
					next_reset_ff <= '1';
					next_hold_ff <= '0';
				when OPERATE =>
						if (dataIn(10) = '0') then
							next_state <= OPERATE;			
							next_reset_ff <= '0';
							next_hold_ff <= '1';
						else                      --no hold, input is valid, ffs are updated
							next_state <= OPERATE;
							next_reset_ff <= '0';
							next_hold_ff <= '0';
						end if;
						
						enable_writing_to_conf <= '0';
						next_configure_count <= 0;
				end case;
			end process;
end arch;
	 


