library IEEE;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;

entity backend is
	generic (
			INPUT_SIZE : natural :=  1;
			NR_STATES  : natural :=  4;
			NR_NFSM    : natural :=  1;
			MEM_SIZE   : natural :=  5;
			MEM_B_SIZE : natural :=  8;
			NUM_REG    : natural := 25
			);
	port (
			clk          : in std_logic;
			in_16        : in std_logic_vector(15 downto 0);
			data_in_sm0  : in std_logic_vector(7 downto 0);
			data_out_sm0 : out std_logic_vector(15 downto 0);		
			
			out_16       : out std_logic_vector(15 downto 0)
		);
end backend;


architecture arch of backend is

	component backend_memory
		generic(
			MEM_B_SIZE   : natural  :=  8;
			NUM_REG      : natural  := 25
		   );
		port (
			clk	       : in std_logic;
			shift_i      : in std_logic_vector (MEM_B_SIZE-1 downto 0);
			write_en     : in std_logic;
			shift_o      : out std_logic_vector (MEM_B_SIZE-1 downto 0)
		);
	end component;
---------------------------SIGNALS FOR STATE MACHINE-------------------------------------
	type operational_state is (IDLE,INIT,CONFIGURE_SM,RUN); 
	signal current_state, next_state             : operational_state;
	signal configure_count, next_configure_count : integer range 0 to NUM_REG+5 := 0;
-----------------------------------------------------------------------------------------

--------------------------SIGNALS FOR BACKEND MEMORY-------------------------------------
	signal shift_in_mem,next_shift_in_mem, shift_out_mem : std_logic_vector (MEM_B_SIZE-1 downto 0);
	signal enable_backend_mem,next_enable_backend_mem   : std_logic;
-----------------------------------------------------------------------------------------

--------------------------INTERNAL SIGNALS-----------------------------------------------
	signal start_configure                    : std_logic;
	signal accept_from_sm                     : std_logic_vector(NR_NFSM-1 downto 0);
	signal accept_input,next_accept_input     : std_logic;
	signal delete_from_sm                     : std_logic_vector(NR_NFSM-1 downto 0);
	signal copy_from_sm                       : std_logic_vector(NR_NFSM-1 downto 0);
	signal sync_flag,next_sync_flag           : std_logic;
	signal output_state_debug                 : std_logic_vector(1 downto 0);
	signal to_configure,next_to_configure     : std_logic_vector(NR_NFSM-1 downto 0); --determines which nfsm to fonfigure, default 01
	signal active_states,next_active_states   : std_logic_vector(NR_NFSM-1 downto 0);-- for output, show the currently active states
-----------------------------------------------------------------------------------------

--------------------------OUTPUTS SIGNALS TO STATE MACHINE(S)----------------------------
	signal reset_sm            : std_logic_vector(NR_NFSM-1 downto 0);
	signal configure_enable_sm : std_logic_vector(NR_NFSM-1 downto 0);
	signal input_sm            : std_logic_vector(7 downto 0);
	signal force_en_sm         : std_logic_vector(NR_NFSM-1 downto 0); 
	signal valid_input         : std_logic;
-----------------------------------------------------------------------------------------

--------------------------INPUT SIGNALS--------------------------------------------------
	--in_16(7 downto 0)  = input vector
	--in_16(8)           = reset
	--in_16(9)          = start configure: on 1, configure first state machine
	--in_16(10)          = sync flag, used to differentiate equal inputs
-----------------------------------------------------------------------------------------

-------------------------OUTPUT SIGNALS--------------------------------------------------
	--out_16(3 downto 0) = accept from different NFSM's
	--out_16(7 downto 4) = active states
-----------------------------------------------------------------------------------------
------------------------SIGNALS FROM SM--------------------------------------------------
	-- dataIn(7-0)    = input from form backend, shifted in
	-- dataIn(8)      = configure enable, 1 active pulse signals configure vektors on next cycle
	-- dataIn(9)      = force states enable signal
	-- dataIN(10)     = valid input
	-- dataIn(11)     = reset, active high
	
	-- dataOut(0)     = accept output from NFSM
	-- dataOut(1)     = copy the state, 
	-- dataOut(2)     = deleted state
-----------------------------------------------------------------------------------------

begin
---------------STATIC CONNECTIONS--------------------------------------------------------
	start_configure         <= in_16(9);
	accept_from_sm(0)       <= data_in_sm0(0);
	copy_from_sm(0)         <= data_in_sm0(1);
	delete_from_sm(0)       <= data_in_sm0(2);

	input_sm                <= in_16(7 downto 0);

	out_16(0)   <= accept_input;
	out_16(1)   <= active_states(0);
	out_16(2)   <= '0';
	out_16(3)   <= '0';
	
	out_16(4)           <= '0';
	out_16(5)           <= '0';
	out_16(6)           <= data_in_sm0(7) or data_in_sm0(3) or data_in_sm0(4) or data_in_sm0(5) or data_in_sm0(6);
	out_16(7)           <= '0';
	
	out_16(9 downto 8)   <= output_state_debug;
	out_16(10)           <= sync_flag;
	out_16(15 downto 11) <= in_16(15 downto 11); 
-----------------------------------------------------------------------------------------

	BACKMEM : backend_memory generic map ( MEM_B_SIZE => MEM_B_SIZE,
										            NUM_REG => NUM_REG
										          )
							port map ( clk      => clk,
										  shift_i  => shift_in_mem,
										  write_en => enable_backend_mem,
										  shift_o  => shift_out_mem
									   );


	CLKPROC : process(clk)
	begin
		if rising_edge(clk) then 
			if(in_16(8) = '1') then
				current_state <= IDLE;
				configure_count <= 0;
				sync_flag <= '0';
				
				data_out_sm0 <= (others=>'0');
				
				output_state_debug  <= (others=>'0');
				enable_backend_mem <= '0';
				shift_in_mem <= (others=>'0');
				to_configure <= (others=>'0');
				active_states <= (others=>'0');
				accept_input <= '0';
			else
				to_configure <= next_to_configure;
				current_state <= next_state;
				configure_count <= next_configure_count;
				sync_flag <= next_sync_flag;
				active_states <= next_active_states;
				accept_input <= next_accept_input;
				
				if ( force_en_sm(0) = '1') then
					data_out_sm0(7 downto 0) <= "00000010";
				elsif ( next_active_states(0) = '1') then
					data_out_sm0(7 downto 0)   <= input_sm;
				else
					data_out_sm0(7 downto 0)   <= next_shift_in_mem;
				end if;
				
				data_out_sm0(8)            <= configure_enable_sm(0);

				data_out_sm0(9)            <= force_en_sm(0);
				
				data_out_sm0(10)				<= valid_input;
				
				data_out_sm0(11)           <= reset_sm(0);

				data_out_sm0(15 downto 12) <= (others=>'0');
				
				enable_backend_mem <= next_enable_backend_mem;
				shift_in_mem <= next_shift_in_mem;
				
				if ( next_state = IDLE) then
					output_state_debug <= "00";
				elsif ( next_state = INIT) then
					output_state_debug <= "01";
				elsif ( next_state = CONFIGURE_SM) then
					output_state_debug <= "10";
				else
					output_state_debug <= "11";
				end if;
				
		   end if;
		end if;
	end process;

	SMPROC : process(start_configure,
						  current_state,
						  configure_count,
						  in_16,
						  sync_flag,
						  shift_out_mem,
						  copy_from_sm,
						  delete_from_sm,
						  accept_from_sm,
						  to_configure,
						  active_states
						  )
	begin
		case current_state is
			when IDLE =>
				if(start_configure = '1') then
					next_state <= INIT;
					next_sync_flag <= '1';
				else
					next_state <= IDLE;
					next_sync_flag <= '0';
				end if;
				
				configure_enable_sm <= (others=>'0');
				reset_sm <= (others=>'1');
				next_enable_backend_mem <= '0';
				next_shift_in_mem <= (others=>'0');
				next_to_configure <= "1";
				next_configure_count <= 0;
				force_en_sm <= (others=>'1');
				valid_input <= '0';
				next_active_states <= (others=>'0');
				next_accept_input  <= '0';
			when INIT =>
			if (configure_count < NUM_REG) then
					if (sync_flag = in_16(10)) then
						next_enable_backend_mem <= '1';
						next_shift_in_mem <= in_16(7 downto 0);
						next_sync_flag <= not sync_flag;
						next_state <= INIT;
						reset_sm <=(others=>'1');
						next_configure_count <= configure_count +1;
					else
						next_enable_backend_mem <= '0';
						next_shift_in_mem <= (others=>'0');
						next_sync_flag <= sync_flag;
						next_state <= INIT;
						reset_sm <=(others=>'1');
						next_configure_count <= configure_count;
					end if;
				else
					next_enable_backend_mem <= '0';
					next_shift_in_mem <= (others=>'0');
					next_sync_flag <= sync_flag;
					next_state <= CONFIGURE_SM;
					reset_sm <=(others=>'0');
					next_configure_count <= 0;
				end if;
				configure_enable_sm <= (others=>'0');
				force_en_sm <= (others=>'0');
				valid_input <= '0';
				next_to_configure <= to_configure;
				next_active_states <= (others=>'0');
				next_accept_input  <= '0';
			when CONFIGURE_SM =>
				if (configure_count = 0) then
					configure_enable_sm <= to_configure;
					next_enable_backend_mem <= '1';
					next_shift_in_mem <= shift_out_mem;
					next_state <= CONFIGURE_SM;
					next_configure_count <= configure_count+1;
					next_to_configure <= to_configure;
					force_en_sm <= (others=>'0');
					next_active_states <= active_states;
				elsif (configure_count < NUM_REG +1) then
					configure_enable_sm <= (others=>'0');
					next_enable_backend_mem <= '1';
					next_shift_in_mem <= shift_out_mem;
					next_state <= CONFIGURE_SM;
					next_configure_count <= configure_count+1;
					next_to_configure <= to_configure;
					force_en_sm <= (others=>'0');
					next_active_states <= active_states;
				elsif (configure_count < NUM_REG+5) then --give some extra time for the configuration of a new NFSM
					configure_enable_sm <= (others=>'0');
					next_enable_backend_mem <= '0';
					next_shift_in_mem <= (others=>'0');
					next_state <= CONFIGURE_SM;
					next_configure_count <= configure_count+1;
					next_to_configure <= to_configure;
					force_en_sm <= (others=>'0');
					next_active_states <= active_states;
				else	
					configure_enable_sm <= (others=>'0');
					next_enable_backend_mem <= '0';
					next_shift_in_mem <= (others=>'0');
					next_state <= RUN;
					next_configure_count <= 0;
					if (to_configure = "1") then --if initial state has been configure
						force_en_sm <= (others=>'0');
						next_active_states <= "1";
					else
						force_en_sm <= "0"; --force clone transition state
						next_active_states <= active_states or to_configure;
					end if;
					next_to_configure <= (others=>'0');
				end if;
			
				reset_sm <= (others=>'0');
				next_accept_input  <= '0';
				next_sync_flag <= sync_flag;
				valid_input <= '0';
			when RUN =>
				if(sync_flag = in_16(10)) then
					next_sync_flag <= not sync_flag;
					valid_input <= '1';
				else
					next_sync_flag <= sync_flag;
					valid_input <= '0';
				end if;
					
				if (copy_from_sm /= "0") then
					next_state <= CONFIGURE_SM;
					next_to_configure <= (others=>'0');
					for i in 0 to NR_NFSM-1 loop
						if (active_states(i) = '0' or delete_from_sm(i) = '1') then
							next_to_configure(i) <= '1';
							exit;
						end if;
					end loop;
				else
					next_state <= RUN;
					next_to_configure <= (others=>'0');
				end if;
				
				if (delete_from_sm /= "0") then --one or more of the active NFSMs is to be deleted/reset
					reset_sm <= delete_from_sm;
					next_active_states <= active_states nand delete_from_sm;
				else
					reset_sm <= (others=>'0');
					next_active_states <= active_states;
				end if;
				
				if ( accept_from_sm /= "0") then --if one of the NFSM's are sendig accept
					next_accept_input <= '1';
				else
					next_accept_input <= '0';
				end if;
				
				force_en_sm <= (others=>'0');

				next_configure_count <= 0;
				configure_enable_sm <= (others=>'0');
				next_enable_backend_mem <= '0';
				next_shift_in_mem <= (others=>'0');
		end case;
	end process;
end arch;




			
