-------------------------------------------------------------------------------------
---------------NFSM module with the states-------------------------------------------
--NFSM clone with the states, as registers -ffs
--Has combinatorial part that evaluates the input and current state, if not on hold
--When there is a valid input, out_states are updatet to new state
--Can signal for accept, copy(clone) or delete
--Accept is sent is last state becomes active
--Delete is sent if no valid transition is found for a valid input
--Copy is sent if a copy(clone) transition is found in the configuration registers
--Module connected paralelly to config mem
--Can be forced to certain state, if singaled to

--author: Tormod Heimark
--date:   2015
--------------------------------------------------------------------------------------

library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;

library work;
use work.check_config_package.all;

entity nfsm is
	 generic (
		  INPUT_SIZE : natural :=  1;
        NUM_REG    : natural := 25;
		  NR_STATES  : natural :=  4;
		  MEM_SIZE   : natural :=  5

	 );
    port (
        input           : in std_logic_vector(INPUT_SIZE-1 downto 0); 
		  clk		         : in std_logic;
		  reset           : in std_logic;
		  hold            : in std_logic;
		  force_states_en : in std_logic;
		  force_states    : in std_logic_vector(NR_STATES-1 downto 0);
		  paralell_in     : in std_logic_vector(NUM_REG*MEM_SIZE-1 downto 0);
					
		  delete_state    : out std_logic;	
		  copy_state      : out std_logic;
		  accept          : out std_logic
    );
end entity;

architecture arch of nfsm is

------------REGISTERS-----------------------
signal config_data_reg_block0: conf_data;
signal config_data_reg_block1: conf_data;
signal config_data_reg_block2 : conf_data;
signal config_data_reg_block3 : conf_data;
signal trans_data : conf_data;
--------------------------------------------

--------OUT OF FLIP FLOPS-------------------
signal out_states : std_logic_vector(NR_STATES-1 downto 0);
-------------------------------------------- 

--------INPUT OF FLIP FLOPS-----------------
signal in_states : std_logic_vector(NR_STATES-1 downto 0);
--------------------------------------------

-------STATUS SIGNALS-----------------------
signal status : conf_info;
signal copy_state_sig, next_copy_state_sig : std_logic;
signal accept_sig : std_logic;
signal delete_state_sig : std_logic;
--------------------------------------------

begin
	status(NR_STATES-1 downto 0) <= out_states;
	status(MEM_SIZE-1 downto NR_STATES) <= input;
	
	
	config_data_reg_block0 <= paralell_in(NUM_REG-1 downto  0);
	config_data_reg_block1 <= paralell_in(2*NUM_REG-1 downto NUM_REG);
	config_data_reg_block2 <= paralell_in(3*NUM_REG-1 downto 2*NUM_REG);
	config_data_reg_block3 <= paralell_in(4*NUM_REG-1 downto 3*NUM_REG);
	trans_data             <= paralell_in(5*NUM_REG-1 downto 4*NUM_REG);
	
	SYNC : process (clk)
		begin
		if rising_edge(clk) then 
			if ( reset = '1' ) then
				out_states(0) <= '1';
				out_states(NR_STATES-1 downto 1) <= (others=>'0');
				accept <= '0';
				copy_state <= '0';
				delete_state <= '0';
			else
				if ( hold = '0') then
					out_states <= in_states;
					copy_state_sig <= next_copy_state_sig;
				else
					if ( force_states_en = '1') then
						out_states <= force_states;
					else
						out_states <= out_states;
					end if;
					copy_state_sig <= '0';
				end if;
				copy_state <= copy_state_sig;
				delete_state <= delete_state_sig;
				accept <= accept_sig;
			end if;
		end if;
	end process;
	
	NFSM0 : process(input,
			  config_data_reg_block0,
			  config_data_reg_block1,
			  config_data_reg_block2,
			  config_data_reg_block3,
			  trans_data,
			  status
			  )
	begin
			in_states(0) <= check_config(config_data_reg_block0,status);
			in_states(1) <= check_config(config_data_reg_block1,status);
			in_states(2) <= check_config(config_data_reg_block2,status);
			in_states(3) <= check_config(config_data_reg_block3,status);
			next_copy_state_sig <= check_config(trans_data,status);
	end process;
	
	DELACC : process (out_states)
			begin
			if (out_states = (out_states'range =>'0')) then
				delete_state_sig <= '1';
				accept_sig <= '0';
			elsif ( out_states(NR_STATES-1) = '1') then
				accept_sig <= '1';
				delete_state_sig <= '0';
			else
				delete_state_sig <= '0';
				accept_sig <= '0';
			end if;

	end process;
end arch;
